clc
clear all

nsamples=12013; %number of times we sample for the Bayes net
addpath(genpath('bnt')) % -- This is required to add the folder 'bnt' with ALL subfolders

N=4; % Number of variables
d=5; %number of edges in the random graph

for i=1:N %Creating variables over which the BN will be defined.
    eval(['A' num2str(i) '= i']);
end
%A1 = 1; A2 = 2; A3 = 3; A4 = 4; A5 = 5; A6 = 6; A7 = 7; A8 = 8; A9= 9; A10 = 10; A11 = 11; A12 = 12; A13 = 13;

prob=rand((N^2-N)/2,1); % generate (N^2-N)/2 -many random numbers
sortedprobsasc=sortrows(prob); %sortrows gives ascending order.
%Now obtain descending order
sortedprobsdes=[];
for uu=1:(N^2-N)/2
sortedprobsdes(uu)=sortedprobsasc((N^2-N)/2+1-uu);
end

%Idea: The d largest numbers stand for an edge in the graph, where the N-d
%smallest numbers represent a missing edge. Hence, we need to transform
%prob into a tuple of d ones and N-d zeros. Divide every number by the
%d(th)-largest number. If the result is less than one, put the number to zero,
%otherwise set it to one.
randomadjanceny=[];
for uu=1:(N^2-N)/2
   randomadjanceny(uu)=sign(floor(prob(uu)/sortedprobsdes(d)));
end

%Insert random tuple of zeros into the adjacency matrix. By default, lower
%numbered variables a parent variables. Hence, there are no cycles.
Graph(1:N,1)=zeros(N,1);
for jj=2:N
    Graph(1:jj-1,jj)=randomadjanceny((((jj-1)^2-(jj-1))/2+1):((jj^2-jj)/2));
    Graph(jj:N,jj)=zeros(N+1-jj,1);
end
    
Parents=zeros(N);
for k=1:N
    if ~isempty(find(Graph(:,k))) % If a variable has a parent, then determine that set of variables
    Parents(1:length(find(Graph(:,k))),k)=find(Graph(:,k));
    end
end

%Next we generate conditional probabilities of variables given their parents.
%For every variable, we first specify the conditional probability of the variable being true, for all states the partents can be in.
CondProb=[];
CondProbs=[];
for qq=1:N
    for pp=1:2^length(find(Graph(:,qq)))
        CondProb(qq,pp)=rand;
    end
    % Next, we need to compute the conditional probability of the negation of the variable, for all states the parents can be in.
    % This is achieved by subtracting the tuple of conditinal probabilities `CondProbs' from 1.
        CondProbs(qq,1:length(CondProb(qq,:)))=CondProb(qq,:);
        CondProbs(qq,length(CondProb(qq,:))+1:2*length(CondProb(qq,:)))=1-CondProb(qq,:);
        CondProb=[];
end

discrete_nodes = 1:N;
node_sizes = 2*ones(1,N); % all variables binary
bnet = mk_bnet(Graph, node_sizes, 'discrete', discrete_nodes);
G = bnet.dag;

% for bb=1:N ---- Fix this loop, if you can!
% bnet.CPD{eval(['A' num2str(i) '= bb'])} = tabular_CPD(bnet, eval(['A' num2str(i) '= bb']), CondProbs(1,1:2^(1+length(find(Graph(:,bb))))));    
% end

%%%%%% By HAND
bnet.CPD{A1} = tabular_CPD(bnet, A1, CondProbs(1,1:2^(1+length(find(Graph(:,1))))));
bnet.CPD{A2} = tabular_CPD(bnet, A2, CondProbs(2,1:2^(1+length(find(Graph(:,2))))));
bnet.CPD{A3} = tabular_CPD(bnet, A3, CondProbs(3,1:2^(1+length(find(Graph(:,3))))));
bnet.CPD{A4} = tabular_CPD(bnet, A4, CondProbs(4,1:2^(1+length(find(Graph(:,4))))));
%bnet.CPD{A5} = tabular_CPD(bnet, A5, CondProbs(5,1:2^(1+length(find(Graph(:,5))))));
%bnet.CPD{A6} = tabular_CPD(bnet, A6, CondProbs(6,1:2^(1+length(find(Graph(:,6))))));
%bnet.CPD{A7} = tabular_CPD(bnet, A7, CondProbs(6,1:2^(1+length(find(Graph(:,7))))));
%bnet.CPD{A8} = tabular_CPD(bnet, A8, CondProbs(6,1:2^(1+length(find(Graph(:,8))))));
%bnet.CPD{A9} = tabular_CPD(bnet, A9, CondProbs(6,1:2^(1+length(find(Graph(:,9))))));
%bnet.CPD{A10} = tabular_CPD(bnet, A10, CondProbs(6,1:2^(1+length(find(Graph(:,10))))));
%bnet.CPD{A11} = tabular_CPD(bnet, A11, CondProbs(6,1:2^(1+length(find(Graph(:,11))))));
%bnet.CPD{A12} = tabular_CPD(bnet, A12, CondProbs(6,1:2^(1+length(find(Graph(:,12))))));
%bnet.CPD{A13} = tabular_CPD(bnet, A13, CondProbs(6,1:2^(1+length(find(Graph(:,13))))));
%%%%%% By HAND


%%% Next, we sample from the BN
pcode mysize  % This is needed to turn "mysize.m" into a .p file
seed = 1; % Do not fix the seed to run simulations, seed=0 makes it repeatable.
rng(seed)
file_name = 'simulation_data';
samples = cell(N, nsamples);
for i = 1:nsamples
    samples(:,i) = sample_bnet(bnet);   % cell array sample{j,i}: 
                                        % value of the jth node in case i
end

% -- This converts the cell structure into a standard number, and takes the
% transpose to make sure observations are in rows, variables in columns
data = cell2num(samples)'-1;

% -- This saves the variable 'data' into the file name chosen above
save(file_name,'data');

%% Next randomly allocate variables to data set with the convention that the center, every appendix and every intersection of two datasets is not empty.
%% First, ensure that these sets are not-empty.
vertices_Dataset1=[1 2];
vertices_Dataset2=[1 3];
vertices_Dataset3=[2 3];
%% Next allocate the remaining variables randomly to these 2^3-1=7 sets.
for xx=4:N
    randomnumber=randi(7);%creates a random integer in {1,2,3,4,5,6,7}.
    if randomnumber==1
        vertices_Dataset1=myunion(vertices_Dataset1,xx);
        vertices_Dataset2=myunion(vertices_Dataset2,xx);
        vertices_Dataset3=myunion(vertices_Dataset3,xx);
    end
    if randomnumber==2
        vertices_Dataset1=myunion(vertices_Dataset1,xx);
        vertices_Dataset2=myunion(vertices_Dataset2,xx);
    end
    if randomnumber==3
        vertices_Dataset1=myunion(vertices_Dataset1,xx);
        vertices_Dataset3=myunion(vertices_Dataset3,xx);
    end
    if randomnumber==4
        vertices_Dataset2=myunion(vertices_Dataset2,xx);
        vertices_Dataset3=myunion(vertices_Dataset3,xx);
    end
    if randomnumber==5
        vertices_Dataset1=myunion(vertices_Dataset1,xx);
    end
    if randomnumber==6
        vertices_Dataset2=myunion(vertices_Dataset2,xx);
    end
    if randomnumber==7
        vertices_Dataset3=myunion(vertices_Dataset3,xx);
    end
    randomnumber=[];
end

L1 = length(vertices_Dataset1); %Number of variables in DataSet1
L2 = length(vertices_Dataset2); %Number of variables in DataSet2
L3 = length(vertices_Dataset3); %Number of variables in DataSet3

% Create Datasets 1, 2, 3

file_name = 'DataSet_One';  % Chose a name for the saved file

%DOne =cell2num(samples)'-1;
data_One = data(:,vertices_Dataset1);


% -- This saves the variable 'data' into the file name chosen above
save(file_name,'data_One');
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


file_name = 'DataSet_Two';  % Chose a name for the saved file

data_Two = data(:,vertices_Dataset2);


% -- This saves the variable 'data' into the file name chosen above
save(file_name,'data_Two');
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

file_name = 'DataSet_Three';  % Chose a name for the saved file

DThree =cell2num(samples)'-1;
data_Three = data(:,vertices_Dataset3);

% -- This saves the variable 'data' into the file name chosen above
save(file_name,'data_Three');

tic


file_name = 'DataSet_One';  % Put the name of the saved file
%--------------------------------------------------------------------------
warning('off','MATLAB:oldPfileVersion')     % Remove Pcode warnings...

% -- This is required to add the folder 'CausalExplorer\Matlab_R14' with
%    ALL subfolders
addpath(genpath('CausalExplorer\Matlab_R14'))

% -- Load your sampled data
load(file_name)

% -- Domain size for the variables is simply the maximum value observed
%    plus one (to allow for 0 valued observations)
%    In our case, {0,1} -> domain_count = 2
domain_counts = max(data_One)+1;

% -- Set up your options for the Causal Explorer.
options.threshold = 0.05;
options.epc = N;%8; % MMHC parameter, which cannot exceed nmber of variables in dataset
options.maxK = N-1;%8; % because it is the upper bound on parents of a variable
options.use_card_lim = 0;
options.max_card = 0;

% -- Call Causal_Explorer on 'data', using Maximum Minimum Hill Climbing
%    The  graph is returned in the sparse matrix 'A'
[A_One,score,stats,time,cps] =...
    Causal_Explorer('MMHC',data_One,domain_counts,'MMHC',options,100,'BDeu');

file_name = 'DataSet_Two';  % Put the name of the saved file
%--------------------------------------------------------------------------
warning('off','MATLAB:oldPfileVersion')     % Remove Pcode warnings...

% -- This is required to add the folder 'CausalExplorer\Matlab_R14' with
%    ALL subfolders
addpath(genpath('CausalExplorer\Matlab_R14'))

% -- Load your sampled data
load(file_name)

% -- Domain size for the variables is simply the maximum value observed
%    plus one (to allow for 0 valued observations)
%    In our case, {0,1} -> domain_count = 2
domain_counts = max(data_Two) + 1;

% -- Set up your options for the Causal Explorer. Play around if you want!
options.threshold = 0.05;
options.epc = N;
options.maxK = N-1;
options.use_card_lim = 0;
options.max_card = 0;

% -- Call Causal_Explorer on 'data', using Maximum Minimum Hill Climbing
%    The  graph is returned in the sparse matrix 'A'
[A_Two,score,stats,time,cps] =...
    Causal_Explorer('MMHC',data_Two,domain_counts,'MMHC',options,100,'BDeu');


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
file_name = 'DataSet_Three';  % Put the name of the saved file


%--------------------------------------------------------------------------
warning('off','MATLAB:oldPfileVersion')     % Remove Pcode warnings...

% -- This is required to add the folder 'CausalExplorer\Matlab_R14' with
%    ALL subfolders
addpath(genpath('CausalExplorer\Matlab_R14'))

% -- Load your sampled data
load(file_name)

% -- Domain size for the variables is simply the maximum value observed
%    plus one (to allow for 0 valued observations)
%    In our case, {0,1} -> domain_count = 2
domain_counts = max(data_Three) + 1;

% -- Set up your options for the Causal Explorer. Play around if you want!
options.threshold = 0.05;
options.epc = N;
options.maxK = N-1;
options.use_card_lim = 0;
options.max_card = 0;

% -- Call Causal_Explorer on 'data', using Maximum Minimum Hill Climbing
%    The  graph is returned in the sparse matrix 'A'
[A_Three,score,stats,time,cps] =...
    Causal_Explorer('MMHC',data_Three,domain_counts,'MMHC',options,100,'BDeu');

Sparse_Overarching_Markov = sparse(N,N); %% the adjancency matrix of the overarching Markov net will
%% be stored here.
for j=1:L1,
              if ~ isempty(find(A_One(:,j)));
                B=find(A_One(:,j));% set of parents of vertex j
                ll=size(find(A_One(:,j)));
                Number_of_Parents_of_j=ll(1,1);% Number of parents of vertex j
                for r=1:Number_of_Parents_of_j-1 % for all pairs of parents
                    for s=r+1:Number_of_Parents_of_j
                        Sparse_Overarching_Markov(vertices_Dataset1(B(r)),vertices_Dataset1(B(s)))=1;
                        % create an edge linking them
                    end
                end
          end
end

%% We want to simply add the adjancecy relation of the first dataset to the edges generated by
%% marrying parents. However, vertices are enumerated as they are in Dataset1. We hence rename them.
[lines, columns, values] = find(A_One);
AAA_One=sparse(vertices_Dataset1(lines)',vertices_Dataset1(columns)',values,N,N);
Sparse_Overarching_Markov=Sparse_Overarching_Markov+AAA_One;

for j=1:L2,
              if ~ isempty(find(A_Two(:,j)));
                B=find(A_Two(:,j));% set of parents of vertex j
                ll=size(find(A_Two(:,j)));
                Number_of_Parents_of_j=ll(1,1);% Number of parents of vertex j
                for r=1:Number_of_Parents_of_j-1 % for all pairs of parents
                    for s=r+1:Number_of_Parents_of_j
                        Sparse_Overarching_Markov(vertices_Dataset2(B(r)),vertices_Dataset2(B(s)))=1;
                        % create an edge linking them
                    end
                end
          end
end

%% We want to simply add the adjancecy relation of the first dataset to the edges generated by
%% marrying parents. However, vertices are enumerated as they are in Dataset2. We hence rename them.
[lines, columns, values] = find(A_Two);
AAA_Two=sparse(vertices_Dataset2(lines)',vertices_Dataset2(columns)',values,N,N);
Sparse_Overarching_Markov=Sparse_Overarching_Markov+AAA_Two;


for j=1:L3,
              if ~ isempty(find(A_Three(:,j)));
                B=find(A_Three(:,j));% set of parents of vertex j
                ll=size(find(A_Three(:,j)));
                Number_of_Parents_of_j=ll(1,1);% Number of parents of vertex j
                for r=1:Number_of_Parents_of_j-1 % for all pairs of parents
                    for s=r+1:Number_of_Parents_of_j
                        Sparse_Overarching_Markov(vertices_Dataset3(B(r)),vertices_Dataset3(B(s)))=1;
                        % create an edge linking them
                    end
                end
          end
end

%% We want to simply add the adjancecy relation of the first dataset to the edges generated by
%% marrying parents. However, vertices are enumerated as they are in Dataset3. We hence rename them.
[lines, columns, values] = find(A_Three);
AAA_Three=sparse(vertices_Dataset3(lines)',vertices_Dataset3(columns)',values,N,N);
Sparse_Overarching_Markov=Sparse_Overarching_Markov+AAA_Three;
% We might have more than one edge between two vertices, so we normalise by
% setting these entries equal to one.
% Also, we need to ensure that our matrix is an upper triangular one.
Symmetric_Sparse_Overarching_Markov=min(Sparse_Overarching_Markov+Sparse_Overarching_Markov',1);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Next we triangulate the overarching Markov network.


%SS=triangulate(full(Symmetric_Sparse_Overarching_Markov), 1:N);


% TRIANGULATE Ensure G is triangulated (chordal), i.e., every cycle of length > 3 has a chord.
% [G, cliques, fill_ins, cliques_containing_node] = triangulate(G, order)
% 
% cliques{i} is the i'th maximal complete subgraph of the triangulated graph.
% fill_ins(i,j) = 1 iff we add a fill-in arc between i and j.
%
% To find the maximal cliques, we save each induced cluster (created by adding connecting
% neighbors) that is not a subset of any previously saved cluster. (A cluster is a complete,
% but not necessarily maximal, set of nodes.)

G=full(Symmetric_Sparse_Overarching_Markov);
order=1:N;
MG = G;
n = length(G);
eliminated = zeros(1,n);
cliques = {};
for i=1:n
  u = order(i);
  U = find(~eliminated); % uneliminated
  nodes = myintersect(neighbors(G,u), U); % look up neighbors in the partially filled-in graph
  nodes = myunion(nodes, u); % the clique will always contain at least u
  G(nodes,nodes) = 1; % make them all connected to each other
  G = setdiag(G,0);  
  eliminated(u) = 1;
  
  exclude = 0;
  for c=1:length(cliques)
    if mysubset(nodes,cliques{c}) % not maximal
      exclude = 1;
      break;
    end
  end
  if ~exclude
    cnum = length(cliques)+1;
    cliques{cnum} = nodes;
  end
end

fill_ins = sparse(triu(max(0, G - MG), 1));

SS=full(G);

%Dau=(SS-full(Symmetric_Sparse_Overarching_Markov))

%Overarching_Markov_Triangulated = view(biograph(triu(SS,1),[],'ShowArrows','off')); % Undirected graph
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Next we want to run a Maximum Cardinality Search
%Such an algorithm has been programmed. However, it only runs properly, if
%the input graph is connected.
%So, we have to first compute the connected components of G.

%%% Compute the Maximum Cardinality Search labelling
%%% Take care that we input only connected components into the mcs-routine
%Since our graph is trinagulated at this point, it has at most N-many cliques
%We will collect the cliques in an array mg, every row is one clique
Final_Bayes_Net=sparse(N,N); % we store the graph of the Bayesian net in this sparse matrix which we
%intialise to an empty sparse N times N matrix


%%%% Changed because I don't have the biostatistics toolbox
ss=sparse(triu(SS,1)); % Turn the adjancency matrix into a sparse matrix
%%   [T,C] = graphconncomp(ss','Directed',false); %T is the number of connected components,
    % the vector C specifies to which connected component a vertex belongs
%%%% END CHANGED
    
%  full(SS)
[T,C] = conncomp(SS);
    
    for x=1:T % for all conneceted components
        verts=find(x==C);%(C<x+1 & C>x-1); %compute the vertices in connected component j
        gg=ss(verts,verts); %restrict adjacency matrix only to those vertices in the j-th connected component
        %full(gg)
        ggg=maxCardinalitySearch(full(gg+gg'),1); %maxCardinalitySearch requires a symmetric adjacency matrix as input and an initial node as input 
        %full(gg+gg')
        %ggg
        %%%%%%%%% Now compute the cliques in each connected component
        %celldisp(maximalCliques(gg,length(C)))
  cc=maximalCliques(gg,length(verts));
    %the second parameter gives an upper bound on the maximal cliques size, which we simply set
   %equal to the size of the connected component of the graph
   
   %%% Next we collect all cliques of a connected component

   if length(verts)>1 %if the connected component only contains one vertex, then there's nothing to do.
        for l=1:length(cc)
              ag=[cc{l}]; %all vertices in the l-th clique of the connected component
                   %% For all elements in ag, we have to give them their number according to Maximum Cardinality Search,
                   %% which we stored in the vector ggg.
                for r=1:length(ag)
                    ag(r)=ggg(ag(r));
                end
                mg(l,1:length(ag))=[ag]; % We now collect all these ``ag''.
        end      
            %% next we want to order the rows of mg lexicographically, i.e., according to Maximum Cardinality Search.
            %% We first want to ignore the entries which are equal to ``0''. We hence replace them all by N+1.
            mg(mg==0)=N+1;
         %% Next we sort each row lexicographically
            sz=sort(mg,2);
        %% Now can sort the rows to obtain an ordering of the cliques which is consistent with Maximum Cardinality Search.
        sorted=sortrows(sz);
        %% Next build the adjacency matrix of the directed graph H
        % In the first clique, we orient by enumeration, i.e., a vertex
        % with a lower MCS number is a parent of a higher numbered vertex
        H=sparse(length(verts),length(verts)); % probably redundant!
        HH=sparse(N,N);
        % Now compute the maximal clique size
        q=size(mg);
        max_clique_size=q(1,2);
        % Let us also store what the variables according to MCS stand for
        True_Names=ones(0,1);
        for i=1:length(verts)
             True_Names(i)=verts(find(ggg==i));
        end
        
        for a=1:max_clique_size-1
            for aa=a+1:max_clique_size
                if sorted(1,a)~=N+1
                    if sorted(1,aa)~=N+1
                       % H(sorted(1,a),sorted(1,aa))=1;
                        HH(True_Names(sorted(1,a)),True_Names(sorted(1,aa)))=1;
                    end
                end
            end
        end
        % Next we have to consider all other cliques in this connected component, if there are any.
if  length(verts)>1       
        D=[]; % intialise the set Di
        for b=2:length(mg(:,1)) %% this used to be length(mg)
            D=myunion(D,sorted(b-1,:)); % These all the vertices in all previous cliques
            E=myintersect(sorted(b,:),D); % These are the vertices in clique b which are in a previously considered clique
            F=setdiff(sorted(b,:),E); % These are the vertices in clique b which were not previously considered
            for c=1:length(E)
                if E(c)~=N+1
                    for d=1:length(F)
                        if F(d)~=N+1
                            %H(E(c),F(d))=1; %vertex c is the parent of vertex d
                            HH(True_Names(E(c)),True_Names(F(d)))=1;
                        end
                     end
                end
            end
        end
end
        mg=[]; % intialise for next connected component
       % H=[];  % intialise for next connected component
        Final_Bayes_Net=Final_Bayes_Net+HH; % For each component, add the edges to the adjacency matrix
        HH=[]; % intialise for next connected component
   end % This ends loop over connected components containing at least two vertices.
    end          
Skeleton=full(Final_Bayes_Net+Final_Bayes_Net')
%vertices_Dataset1
%vertices_Dataset2
%vertices_Dataset3

%% We now compute conditional probabilities
Problematic_vertices=zeros(0,1); % intiliase set of problematic vertices,
%i.e., all those variables for which there does not exist a dataset which
%measured it and all its parents.
  for a=1:N %% for all variables in the Bayesian net

      %% To simplify the computation we consider cases,
      %% if a is a root vertex, then there is no need to determine the parents of a
    if isempty(find(Final_Bayes_Net(:,a))); % If a variable A does not have a parent, i.e., it is a
            %root of a connected component, then we need to find the
            %probability of A and A. Since P(A)+P(A)=1, we only have to
            %compute one of these probabilities.       
            % Since A is a variable in our domain of interest, it is measured
            % in at least one dataset.
            %a
            if  ismember(a,vertices_Dataset1)
                name_of_a=find(vertices_Dataset1==a); %find the name of the variable A in Dataset1
                % Next compute for how many rows in data_One the a'th entry
                % is equal to one.
             c=length(find(data_One(:,name_of_a))==1)/nsamples;
             d=1-length(find(data_One(:,name_of_a))==1)/nsamples;
            %% We shall use a matrix Y to collect all conditional probabilities of vertices which are
            %% not problematic colliders.
             Y(a,1:2)=[d,c];
            elseif ismember(a,vertices_Dataset2)
             name_of_a=find(vertices_Dataset2==a);%find the name of the variable A in Dataset2
             c=length(find(data_Two(:,name_of_a))==1)/nsamples;
             d=1-length(find(data_Two(:,name_of_a))==1)/nsamples;
             Y(a,1:2)=[d,c];
            elseif ismember(a,vertices_Dataset3)
             name_of_a=find(vertices_Dataset3==a);%find the name of the variable A in Dataset3
             c=length(find(data_Three(:,name_of_a))==1)/nsamples;
             d=1-length(find(data_Three(:,name_of_a))==1)/nsamples;
             Y(a,1:2)=[d,c];
            end
    end

%% The following loop could be sped up.
%% As it is, we check whether a line in a dataset corresponds to a state. We do so for many states.
%% Since we know that every line correspond to some state, it would be quicker to compute for every line the corresponding state and increase a counter for every state. FUTURE WORK.
    
%% If A_a is not the root node, then we have to compute the parents of A_a
    if ~isempty(find(Final_Bayes_Net(:,a)));
        parents_of_a=find(Final_Bayes_Net(:,a));%Note that this tuple is ordered.
        a_union_parents_of_a=myunion(a,parents_of_a);%And so is this tuple because of the myunion command.
        % Now we have to determine whether a is a problematic collider
        % if it is not, then we can determine the y-parameters directly
        % from a dataset
        if all(ismember(a_union_parents_of_a, vertices_Dataset1));% this checkes whether A_a and all of
            %its parents are measured in Dataset1. If so, then (since
            %datasets are assumed to be consistent) we can use Dataset1 to
            %determine the relevant conditional probabilities.
             for mm=1:length(parents_of_a)
                names_of_parents_of_a(mm)=find(vertices_Dataset1==parents_of_a(mm));
             end
         name_of_a=find(vertices_Dataset1==a);
         names_of_a_and_parents=myunion(name_of_a,names_of_parents_of_a);
         place_of_a=find(names_of_a_and_parents==name_of_a);
            %Next we compute the names of the parents of a
            States_generated_by_parents_of_a=getcondvects(length(parents_of_a));
            % For each state of Dataset1 which instantiates such a substate we determine which fraction
            % also instantiates A, respectively, A.
            for r=1:2^(length(parents_of_a)) %for every substate generated by the parents of a
                           counter=0;
                           counter_2=0;
%                           counter_3=0;
                for nn=1:nsamples % we check every line of Dataset1
                    line_nn_of_DataSet=data_One(nn,:);
                    if States_generated_by_parents_of_a(r,:)==line_nn_of_DataSet(names_of_parents_of_a); % we check
                        %whether this line agrees with the substate number r
                        %generated by the parents -- the first substate is all zeros.
                        counter=counter+1;
                        if 1== line_nn_of_DataSet(name_of_a);
                            counter_2=counter_2+1;
                        end
                        %% Thisloop is only for checking correctness of the algorithm
%                         if 0== nng(find(vertices_Dataset1==a));
%                           counter_3=counter_3+1;
%                         end
%% End checking loop
                    end
                end
                
% For every fixed state generated by the parents of A_a we need to add "A_a"
% and "~A_a" as conjuncts. This generates a state on 1+length(parents_of_a)
% many variables. To compute the number of such a state we think in terms
% of binary numbers.
%We have to insert A_a into set of its parents at the right place.
                UU=1;%intialise
                for u=1:-1+place_of_a
                    UU=UU+2^(length(parents_of_a)-u+1)*(States_generated_by_parents_of_a(r,u));
                end
                for u=place_of_a:length(parents_of_a)
                    UU=UU+2^(length(parents_of_a)-u)*(States_generated_by_parents_of_a(r,u));
                end
                %If A_a occurrs negated we do not have to add anything
                Shifted_UU_for_positive_a=UU+2^(1+length(parents_of_a)-place_of_a);
                %If A_a occurrs positively, then we have to add 2 to the
                %power where the name of A_a appears in a_union_parents_of_a
              if counter>0
              Y(a,UU)=(counter-counter_2)/counter;
              Y(a,Shifted_UU_for_positive_a)=counter_2/counter;
                end
                if counter==0
              Y(a,UU)=0;
              Y(a,Shifted_UU_for_positive_a)=0;
                end
%The row Y(a,:) consists of the conditional probabilities at A_a;
%enumerated by the valuations on a_union_parents_of_a.


              %Y(a,2*r-1)=counter_2/counter;
              %Y(a,2*r)=(counter-counter_2)/counter;
              % number of lines of Dataset1 instantiate substate r and
              % which fraction of these lines instantiates A and A.
            end
            names_of_parents_of_a=0; % initiliase for next loops
        elseif all(ismember(a_union_parents_of_a, vertices_Dataset2))
                         for mm=1:length(parents_of_a)
                names_of_parents_of_a(mm)=find(vertices_Dataset2==parents_of_a(mm));
                         end
            %Next we compute the names of the parents of a
name_of_a=find(vertices_Dataset2==a);
names_of_a_and_parents=myunion(name_of_a,names_of_parents_of_a);
place_of_a=find(names_of_a_and_parents==name_of_a);
States_generated_by_parents_of_a=getcondvects(length(parents_of_a));
                         for r=1:2^(length(parents_of_a)) %for every substate generated by the parents of a
                            counter=0;
                            counter_2=0;
                             for nn=1:nsamples % we check every line of Dataset2
                                 line_nn_of_DataSet=data_Two(nn,:); % call such a line nng
                                     if States_generated_by_parents_of_a(r,:)==line_nn_of_DataSet(names_of_parents_of_a);
                        % we check
                        %whether this line agrees with the substate number r
                        %generated by the parents -- the first substate is all zeros.
                                         counter=counter+1;
                                        if 1== line_nn_of_DataSet(find(vertices_Dataset2==a));
                                            counter_2=counter_2+1;
                                        end
                                    end
                             end
                             UU=1;%intialise
                for u=1:-1+place_of_a
                    UU=UU+2^(length(parents_of_a)-u+1)*(States_generated_by_parents_of_a(r,u));
                end
                for u=place_of_a:length(parents_of_a)
                    UU=UU+2^(length(parents_of_a)-u)*(States_generated_by_parents_of_a(r,u));
                end
                %If A_a occurrs negated we do not have to add anything
                Shifted_UU_for_positive_a=UU+2^(1+length(parents_of_a)-place_of_a);
                %If A_a occurrs positively, then we have to add 2 to the
                %power where the name of A_a appears in a_union_parents_of_a
              if counter>0
              Y(a,UU)=(counter-counter_2)/counter;
              Y(a,Shifted_UU_for_positive_a)=counter_2/counter;
                end
                if counter==0
              Y(a,UU)=0;
              Y(a,Shifted_UU_for_positive_a)=0;
                end              
%Y(a,2*r-1)=counter_2/counter;
%Y(a,2*r)=(counter-counter_2)/counter;
              % how number of lines of Dataset2 instantiate substate r and
              % how which fraction of these lines instantiates A and A.
                         end
                        names_of_parents_of_a=0; % initiliase for next loops
        elseif all(ismember(a_union_parents_of_a, vertices_Dataset3))
                names_of_parents_of_a=0; % initiliase for next loops
                         for mm=1:length(parents_of_a)
                names_of_parents_of_a(mm)=find(vertices_Dataset3==parents_of_a(mm));
                         end
name_of_a=find(vertices_Dataset3==a);
names_of_a_and_parents=myunion(name_of_a,names_of_parents_of_a);
place_of_a=find(names_of_a_and_parents==name_of_a);
            %Next we compute the names of the parents of a
            States_generated_by_parents_of_a=getcondvects(length(parents_of_a));
                         for r=1:2^(length(parents_of_a)) %for every substate generated by the parents of a
                            counter=0;
                            counter_2=0;
                             for nn=1:nsamples % we check every line of Dataset3
                                 line_nn_of_DataSet=data_Three(nn,:); % call such a line nng
                                     if States_generated_by_parents_of_a(r,:)==line_nn_of_DataSet(names_of_parents_of_a);
                        % we check
                        %whether this line agrees with the substate number r
                        %generated by the parents -- the first substate is all zeros.
                                         counter=counter+1;
                                        if 1== line_nn_of_DataSet(find(vertices_Dataset3==a));
                                            counter_2=counter_2+1;
                                        end
                                    end
                             end
                             UU=1;%intialise
                for u=1:-1+place_of_a
                    UU=UU+2^(length(parents_of_a)-u+1)*(States_generated_by_parents_of_a(r,u));
                end
                for u=place_of_a:length(parents_of_a)
                    UU=UU+2^(length(parents_of_a)-u)*(States_generated_by_parents_of_a(r,u));
                end
                %If A_a occurrs negated we do not have to add anything
                Shifted_UU_for_positive_a=UU+2^(1+length(parents_of_a)-place_of_a);
                %If A_a occurrs positively, then we have to add 2 to the
                %power where the name of A_a appears in a_union_parents_of_a
                if counter>0
              Y(a,UU)=(counter-counter_2)/counter;
              Y(a,Shifted_UU_for_positive_a)=counter_2/counter;
                end
                if counter==0
              Y(a,UU)=0;
              Y(a,Shifted_UU_for_positive_a)=0;
                end
%Y(a,2*r-1)=counter_2/counter;
%Y(a,2*r)=(counter-counter_2)/counter;
              % how number of lines of Dataset2 instantiate substate r and
              % how which fraction of these lines instantiates A and A.
                         end
         names_of_parents_of_a=0; % initiliase for next loops
                
        elseif 0==0
            %a 
            % This loop makes the
            % lines corresponding to under-determined variables stand out
            % in the conditional probability matrix.
        Number_of_states_generated_by_parents_of_a_and_a=2^(length(parents_of_a)+1);
        Y(a,1)=length(parents_of_a);
        Y(a,2:length(parents_of_a)+1)=parents_of_a;
        Problematic_vertices=myunion(a,Problematic_vertices);
            names_of_parents_of_a=0; % initiliase for next loops
        end
    end
  end
      Y
 
 %%% IF(!!!) there is no optimisation problem to be solved, then we are done here.     
      if isempty(Problematic_vertices)
          toc
      end
      
 %Problematic_vertices
 
 
 if ~isempty(Problematic_vertices)
 Problematic_vertices= setdiff(Problematic_vertices,[])%This ensures that the tuple is well-ordered 
%% We now compute the number of unknown y-parameters
if ~isempty(Problematic_vertices) % if there exists at least one such y-parameter
    Number_of_y_paramters=0; %initialise
    for q=1:length(Problematic_vertices) % for every problematic vertex
Number_of_y_paramters=Number_of_y_paramters+2^(length(find(Final_Bayes_Net(:,Problematic_vertices(q)))));
% compute the number of parents. Raise this number to the second power.
% This then counts the number of conditionals of the form P(A|State of
% parents of A). Obtain P(~A|State of parents of A)=1-P(A|State of
% parents of A). All these numbers of y-parameters need to be added.
    end
end
if isempty(Problematic_vertices)
    Number_of_y_paramters=0;
end
   Number_of_y_paramters

   
%%%% THIS NEEDS TO BE TESTED FOR MULTIPLE UNDERDETERMINED VARIABLES   
%Let us now compute the number of y-parameters coming prior to the some under-determined variable
if ~isempty(Problematic_vertices)
    Number_of_y_parameters_prior(1)=0;
    Sum_Number_of_y_parameters_prior(1)=0;
    for PVV=2:length(Problematic_vertices)
        Number_of_y_parameters_prior(PVV)=2^(length(find(Final_Bayes_Net(:,Problematic_vertices(PVV-1)))));
        Sum_Number_of_y_parameters_prior(PVV)=Sum_Number_of_y_parameters_prior(PVV-1)+Number_of_y_parameters_prior(PVV);
    end
end

X0 =(0.5)*ones(1,Number_of_y_paramters);
% % Make a starting guess at the solution -- One entry for every variable!
% %Starting guess is: all conditional probabilities equal to one half.
 options = optimset('Algorithm','interior-point' ,'TolFun', 1e-3,'TolX', 1e-3, 'TolCon',1e-6,'MaxFunEvals',10000000);%,'TolCon',1e-16, 'Display', 'iter-detailed','interior-point'
% options = optimset('Algorithm','sqp', 'MaxIter', 1000000);
 


[X,fval] = ... 
 fmincon(@(X)Objective(X,Number_of_y_paramters,data_One,data_Two,data_Three,nsamples,Y,vertices_Dataset1,...
    vertices_Dataset2,vertices_Dataset3,N,Problematic_vertices,Final_Bayes_Net,Sum_Number_of_y_parameters_prior),...
        X0,[eye(Number_of_y_paramters);-eye(Number_of_y_paramters)],...
        [ones(1,Number_of_y_paramters)';zeros(1,Number_of_y_paramters)'],[],[],[],[],...
@(X)constraintsyNEW(X,data_One,data_Two,data_Three,nsamples,Y,vertices_Dataset1,vertices_Dataset2,vertices_Dataset3,N,Problematic_vertices,Final_Bayes_Net,Sum_Number_of_y_parameters_prior),...
    options);
OurSolution=X
% 
% 
% % v = 0.3:0.3:0.6;
% % [X,Y,Z,T] = meshgrid(v);
% % ptmatrix = [X(:),Y(:),Z(:),T(:)];
% % tpoints = CustomStartPointSet(ptmatrix);
% % ms = MultiStart;
% % 
% % 
% % opts = optimoptions(@fmincon,'Algorithm','interior-point');
% % problem = createOptimProblem('fmincon','objective',@(X)Objective(X,Number_of_y_paramters,data_One,data_Two,data_Three,...
% %     nsamples,Y,vertices_Dataset1,vertices_Dataset2,vertices_Dataset3,N,Problematic_vertices,Final_Bayes_Net,...
% %     Sum_Number_of_y_parameters_prior),'x0',X0,...
% %     'Aineq',[eye(Number_of_y_paramters);-eye(Number_of_y_paramters)],...
% %     'bineq',[ones(1,Number_of_y_paramters)';zeros(1,Number_of_y_paramters)'],...
% %     'nonlcon', @(X)constraintsyNEW(X,data_One,data_Two,data_Three,nsamples,Y,vertices_Dataset1,vertices_Dataset2,...
% %         vertices_Dataset3,N,Problematic_vertices,Final_Bayes_Net,Sum_Number_of_y_parameters_prior),'options',opts);
% %   %  gs = GlobalSearch(ms);
% %     x = run(ms,problem,tpoints)
% % %[x,fval] = run(gs,problem)
% % OurSolution=X
% Entropy_of_OurSolution=Objective(OurSolution,Number_of_y_paramters,data_One,data_Two,data_Three,nsamples,Y,vertices_Dataset1,vertices_Dataset2,vertices_Dataset3,N,Problematic_vertices,Final_Bayes_Net,Sum_Number_of_y_parameters_prior) 

toc
 end

 
% Test=[Res(1)/(Res(1)+Res(2));Res(3)/(Res(3)+Res(4));Res(5)/(Res(5)+Res(6));Res(7)/(Res(7)+Res(8))]
% constraintsyNEW(Test,data_One,data_Two,data_Three,nsamples,Y,vertices_Dataset1,vertices_Dataset2,vertices_Dataset3,N,Problematic_vertices,Final_Bayes_Net,Sum_Number_of_y_parameters_prior)
% Objective(Test,Number_of_y_paramters,data_One,data_Two,data_Three,nsamples,Y,vertices_Dataset1,vertices_Dataset2,vertices_Dataset3,N,Problematic_vertices,Final_Bayes_Net,Sum_Number_of_y_parameters_prior)
% Objective(OurSolution,Number_of_y_paramters,data_One,data_Two,data_Three,nsamples,Y,vertices_Dataset1,vertices_Dataset2,vertices_Dataset3,N,Problematic_vertices,Final_Bayes_Net,Sum_Number_of_y_parameters_prior)
 
 
% Number_of_y_parameters_prior
% Sum_Number_of_y_parameters_prior
% Final_Bayes_Net
% vertices_Dataset1
% vertices_Dataset2
% vertices_Dataset3