function [c, Constraints] = constraintsyNEW(X,data_One,data_Two,data_Three,nsamples,Y,vertices_Dataset1,vertices_Dataset2,vertices_Dataset3,N,Problematic_vertices,Final_Bayes_Net,Sum_Number_of_y_parameters_prior)

% Nonlinear inequality constraints
c = []; %There are no inequality constraints computed here. They are taken care of in the calling routine.
% There are equality constraints to be computed.
Constraints=[];

for i=1:N %case distinction of whether problematic vertex or not
    if ismember(i,Problematic_vertices) %If A_i is a problematic vertex
P_V=i;
Constraints_at_PV=[];
Parents_of_PV=find(Final_Bayes_Net(:,P_V));
PV_and_Parents=myunion(P_V,Parents_of_PV);%this is ordered, myunion does this for us
Powerset=powerset(PV_and_Parents);
LeftHandSide_of_Constraint_At_PV=[];
Constraints_RHS_at_PV=[];
% First, take the union of a problematic vertex and its parents. Then, compute the
% powerset of this set.

% Constraints arise for each set of vertices in this powerset which are
% measured together. However, if there exists a superset of vertices which
% are also jointly measured, then the constraints arising from the superset
% entail the constraints arising from the smaller set.
% Since the vertex is problematic, we can also know that the vertex and all
% its parents are not jointly measured. The entire power set does hence not give rise to constraints.
upper_bound=2^(length(find(Final_Bayes_Net(:,P_V)))+1)-1;
candidates=[];
candidates=2:upper_bound; % list all elements of the powerset which are potentially maximal sets of
% jointly measured sets of variables. Powerset{1} is the empty set of
% vertices, which we can, of course, ignore.

% Note that only datasets which measure the under-determined variable
% give rise to constraints which concern us here. We can hence ignore all
% subsets of variables which do not contain the under-determined variable and
% hence delete them from the candidate set.

        for u=2:upper_bound
            if ismember(P_V,Powerset{u})==0
            candidates=setdiff(candidates,u);
            end
        end

        for j=1:upper_bound-1% up to the number of possible sets of vertices which are
        %maximal subsets. The powerset contains 2^(Y(P_V,1)+1)-many elements.
        % We can ignore the empty set and the entire set. 

            if all(ismember(Powerset{2^(Y(P_V,1)+1)-j},vertices_Dataset1))==1;

            %% If we have found a dataset which jointly measures a set of vertices, and no other
            % dataset measures a superset of these vertices, then there are two things to
            % compute. Since we start by investigating the elements of the
            % powerset with most elements, no superset of vertices can be jointly measured in a
            % dataset.
            %
            % First, all other sets of vertices which are subsets of this set can now be ignored.
            % Second, constraints are arising which need to be computed.
            % We shall first delete these subsets from the candidates list.
            % We need to delete all candidates for which Powerset{l} is a
            % subset of Powerset{2^(Y(P_V,1)+1)-j}.
                for l=2:2^(Y(P_V,1)+1)-j-1
                    if ismember(l,candidates) % for all candidate subsets
                        if all(ismember(Powerset{l},Powerset{2^(Y(P_V,1)+1)-j})) %which are subsets of Powerset{i}
                        candidates=setdiff(candidates,l); %remove these sets from the candidates list
                        end
                    end
                end
            
            elseif all(ismember(Powerset{2^(Y(P_V,1)+1)-j},vertices_Dataset2))==1;
            %% If we have found a dataset which jointly measures a set of vertices, and no other
            % dataset measures a superset of these vertices, then there are two things to
            % compute. Since we start by investigating the elements of the
            % powerset with most elements, no superset of vertices can be jointly measured in a
            % dataset.
            %
            % First, all other sets of vertices which are subsets of this set can now be ignored.
            % Second, constraints are arising which need to be computed.
            % We shall first delete these subsets from the candidates list.
            % We need to delete all candidates for which Powerset{l} is a
            % subset of Powerset{2^(Y(P_V,1)+1)-j}.
                for l=2:2^(Y(P_V,1)+1)-j-1
                    if ismember(l,candidates) % for all candidate subsets
                        if all(ismember(Powerset{l},Powerset{2^(Y(P_V,1)+1)-j})) %which are subsets of Powerset{i}
                        candidates=setdiff(candidates,l); %remove these sets from the candidates list
                        end
                    end
                end
            elseif all(ismember(Powerset{2^(Y(P_V,1)+1)-j},vertices_Dataset3))==1;
            %% If we have found a dataset which jointly measures a set of vertices, and no other
            % dataset measures a superset of these vertices, then there are two things to
            % compute. Since we start by investigating the elements of the
            % powerset with most elements, no superset of vertices can be jointly measured in a
            % dataset.
            %
            % First, all other sets of vertices which are subsets of this set can now be ignored.
            % Second, constraints are arising which need to be computed.
            % We shall first delete these subsets from the candidates list.
            % We need to delete all candidates for which Powerset{l} is a
            % subset of Powerset{2^(Y(P_V,1)+1)-j}.
                for l=2:2^(Y(P_V,1)+1)-j-1
                    if ismember(l,candidates) % for all candidate subsets
                        if all(ismember(Powerset{l},Powerset{2^(Y(P_V,1)+1)-j})) %which are subsets of Powerset{i}
                            candidates=setdiff(candidates,l); %remove these sets from the candidates list
                        end
                    end
                end
            else
                candidates=setdiff(candidates,2^(Y(P_V,1)+1)-j);
            end
        end
    % We now collect all candidate sets which have been jointly measured.
    Maximal_Jointly_Observed_Variables=[];
        for r=1:length(candidates)
            Maximal_Jointly_Observed_Variables(r,1:length(Powerset{candidates(r)}))=Powerset{candidates(r)};
            
        end

       
        for g=1:length(Maximal_Jointly_Observed_Variables(:,1)) % for every such set, constraints
            %have to be satisfied. We shall now compute these constraints.
            PV_AND_Measured_Parents=[];
            PV_AND_Measured_Parents=Maximal_Jointly_Observed_Variables(g,:);
            
            if ismember(0,Maximal_Jointly_Observed_Variables(g,:))
                %Maximal_Jointly_Observed_Variables
            PV_AND_Measured_Parents=Maximal_Jointly_Observed_Variables(g,1:min(find(Maximal_Jointly_Observed_Variables(g,:)==0))-1);
            %SomethingGoneWrong
            end
            if ~ismember(0,Maximal_Jointly_Observed_Variables(g,:))
            PV_AND_Measured_Parents=Maximal_Jointly_Observed_Variables(g,:);
            end
            
            if ismember(0,PV_AND_Measured_Parents)
                17
                PV_AND_Measured_Parents
                Maximal_Jointly_Observed_Variables
                SomethingGoneWrong
            end
                
            Measured_Parents_of_PV=setdiff(PV_AND_Measured_Parents,P_V);
            Contstraints_RHS_AT_PV_Max_Observed_Vars=zeros(2^(length(PV_AND_Measured_Parents)),1);
            %Measured_Parents=setdiff(PV_AND_Measured_Parents,P_V);
            Other_Ancestors_of_PV=setdiff(ancestors(Final_Bayes_Net,i),PV_AND_Measured_Parents);
            Other_Parents_of_PV=setdiff(find(Final_Bayes_Net(:,P_V)),PV_AND_Measured_Parents);
            Name_of_PV_in_Measured_Parents=find(PV_AND_Measured_Parents==P_V);
            for v=1:2^(length(PV_AND_Measured_Parents))%For every valuation over these variables
                % Let us first turn the number into a 0-1 string
                binary_of_v=dec2bin(2^(length(PV_AND_Measured_Parents))+v-1); 
                for u=1:length(PV_AND_Measured_Parents)
                    rrv=num2str(binary_of_v);
                    qqqv(u)=rrv(u+1);
                end
               
                for hh=1:length(qqqv)
                    Binary_eval_of_v(hh)=str2num(qqqv(hh));
                end % Ends loop to turn valuation v into a binary number
                %Binary_eval_of_v
                sum=0;
                Binary_eval_of_s=[];
                for s=1:2^(length(Other_Ancestors_of_PV))
             %For every valuation over these ancestors of PV
%variables without these variables, we now have to first determine the
%y-parameter.

%Question: Which is the superscript of y_i^?
%Answer: The superscript is the valuation of the parents, which consists of the
%union of the valuation v and the part of the valuation s which assigns
%truth values to the parents not assigned a truth value by v.
            % Let us first turn the number into a 0-1 string
                    Binary_eval_of_s=[];
                    binary_of_s=dec2bin(2^(length(Other_Ancestors_of_PV))+s-1); 
                        for u=1:length(Other_Ancestors_of_PV)
                            rrs=num2str(binary_of_s);
                            qqqs(u)=rrs(u+1);
                        end
                        for hhs=1:length(qqqs)
                            Binary_eval_of_s(hhs)=str2num(qqqs(hhs));
                        end % Ends loop to turn valuation s into a binary number
                            %Binary_eval_of_s
                        VAL=1; %initialises the valuation of the parents of PV
                        for g=1:length(Parents_of_PV)
                            if ismember(Parents_of_PV(g),PV_AND_Measured_Parents)          
VAL=VAL+2.^(length(Parents_of_PV)-g).*Binary_eval_of_v(find(PV_AND_Measured_Parents==Parents_of_PV(g)));
%Determine which valuation, v or s, valuates the g-th parent of PV.
                            end
                            if ismember(Parents_of_PV(g),Other_Parents_of_PV)
VAL=VAL+2.^(length(Parents_of_PV)-g).*Binary_eval_of_s(find(Other_Ancestors_of_PV==Parents_of_PV(g)));
                            end
                        end
%Next determine, whether we have to use the y-parameter or one minus the
%y-parameter.

    Number_of_PV=find(Problematic_vertices==P_V);
   if Binary_eval_of_v(find(PV_AND_Measured_Parents==P_V))==0
    prod=X(Sum_Number_of_y_parameters_prior(Number_of_PV)+VAL);    
   end
   if Binary_eval_of_v(find(PV_AND_Measured_Parents==P_V))==1
    prod=1-X(Sum_Number_of_y_parameters_prior(Number_of_PV)+VAL);
   end

                        %We now multiply the conditional probabilities at the
%ancestors of P_V. We have to multiply with a factor for every ancestor of
%P_V. Which exact conditional is determined by the valuations v and s.
Ancestors=ancestors(Final_Bayes_Net,P_V);
                        for q=1:length(Ancestors)
Ancestor_q=Ancestors(q);
if ~ismember(Ancestor_q,Problematic_vertices)%if the vertex is non-problematic, then we look up
    %the factor in the Y-matrix
Parents_of_Ancestor_q=find(Final_Bayes_Net(:,Ancestor_q));
Parents_of_Ancestor_q_AND_Ancestor_q=myunion(Parents_of_Ancestor_q,Ancestor_q);
VAL_at_Ancestor_q=1; %initialising
                            for f=1:length(Parents_of_Ancestor_q_AND_Ancestor_q)
%for all parents of Ancestor_q, as well as Ancestor_q itself, we now have
%to determine the value of the valuation v, respectively, s.
                                if ismember(Parents_of_Ancestor_q_AND_Ancestor_q(f),PV_AND_Measured_Parents)
VAL_at_Ancestor_q=VAL_at_Ancestor_q+(2.^(length(Parents_of_Ancestor_q_AND_Ancestor_q)-f).*Binary_eval_of_v(find(PV_AND_Measured_Parents==Parents_of_Ancestor_q_AND_Ancestor_q(f))));

                                else
VAL_at_Ancestor_q=VAL_at_Ancestor_q+(2.^(length(Parents_of_Ancestor_q_AND_Ancestor_q)-f).*Binary_eval_of_s(find(Other_Ancestors_of_PV==Parents_of_Ancestor_q_AND_Ancestor_q(f))));
                              end
                            end
                            Y(Ancestor_q,VAL_at_Ancestor_q);
                           prod=prod.*Y(Ancestor_q,VAL_at_Ancestor_q);
                        end
                        
if ismember(Ancestor_q,Problematic_vertices)%if the vertex is problematic, then we use y-parameters
%% VAL_at_Ancestor_q needs to be restricted the parents of P_V=Ancestor_q
VAL_at_Ancestor_q_without_q=1; %initialising
             for ff=1:length(Parents_of_Ancestor_q)
                    if ismember(Parents_of_Ancestor_q(ff),PV_AND_Measured_Parents)
VAL_at_Ancestor_q_without_q=VAL_at_Ancestor_q_without_q+(2.^(length(Parents_of_Ancestor_q)-ff).*Binary_eval_of_v(find(PV_AND_Measured_Parents==Parents_of_Ancestor_q(ff))));
                    else
VAL_at_Ancestor_q_without_q=VAL_at_Ancestor_q_without_q+(2.^(length(Parents_of_Ancestor_q)-ff).*Binary_eval_of_s(find(Other_Ancestors_of_PV==Parents_of_Ancestor_q(ff))));
                    end
      end

    if ismember(Ancestor_q,PV_AND_Measured_Parents)%We need to determine the valuation at Ancestor_q
%For this we need to determine, whether Ancestor_q is a member of
%PV_AND_MEasured_Parents. If it is, then the valuation of Ancestor_q is
%determined by the valuation v, otherwise it's determined by valuation s.
Number_of_PV=find(Problematic_vertices==Ancestor_q);
        if Binary_eval_of_v(find(PV_AND_Measured_Parents==Ancestor_q))==0
% as usual, it the valuation we use the y-paramter. If the valuation equals
% one we use one minus the y-parameter.
        prod=prod.*X(Sum_Number_of_y_parameters_prior(Number_of_PV)+VAL_at_Ancestor_q_without_q);
        end
    end
    if ~ismember(Ancestor_q,PV_AND_Measured_Parents)
        if Binary_eval_of_s(find(Other_Ancestors_of_PV==Ancestor_q))==0
        prod=prod.*X(Sum_Number_of_y_parameters_prior(Number_of_PV)+VAL_at_Ancestor_q_without_q);
        end
    end
    if ismember(Ancestor_q,PV_AND_Measured_Parents)
        if Binary_eval_of_v(find(PV_AND_Measured_Parents==Ancestor_q))==1
        prod=prod.*(1-X(Sum_Number_of_y_parameters_prior(Number_of_PV)+VAL_at_Ancestor_q_without_q));
        end
    end
    if ~ismember(Ancestor_q,PV_AND_Measured_Parents)
        if Binary_eval_of_s(find(Other_Ancestors_of_PV==Ancestor_q))==1
        prod=prod.*(1-X(Sum_Number_of_y_parameters_prior(Number_of_PV)+VAL_at_Ancestor_q_without_q));
        end
    end
end
           end             
             sum=sum+prod;
             %Binary_eval_of_v
             %Binary_eval_of_s
                end%ends loop over valuation s
                LeftHandSide_of_Constraint_At_PV=[LeftHandSide_of_Constraint_At_PV; sum];
            end %ends loop over the valuation v of PV and maximal set of jointly observed parents
            
            %% We now compute the right hand side of the constraint at PV
% First, we have to determine the ``names'' of PV and its measured parents
% in a/the dataset in which they are jointly measured
                Names_of_PV_AND_Measured_Parents=[];
                Constraints_RHS_AT_PV_Max_Observed_Vars=zeros(2^(length(PV_AND_Measured_Parents)),1);
                if all(ismember(PV_AND_Measured_Parents,vertices_Dataset1))==1
                    for qg=1:length(PV_AND_Measured_Parents)
                        Names_of_PV_AND_Measured_Parents(qg)=find(vertices_Dataset1==PV_AND_Measured_Parents(qg));
                    end
                    
%Right_Order_Names_of_Measured_Parents_and_PV=setdiff(Names_of_PV_AND_Measured_Parents,P_V); 
                        for nn=1:nsamples % we check every line of Dataset1                                                     
                                relevant_Part_Of_Line_nn=data_One(nn,Names_of_PV_AND_Measured_Parents);
                            kk=1;
                                for nnn=1:length(PV_AND_Measured_Parents)%every partial line corresponds to some state. We next determine this state, called kk. 
kk=kk+2^(length(PV_AND_Measured_Parents)-nnn).*relevant_Part_Of_Line_nn(nnn);
                                end
                                %[PV_AND_Measured_Parents ,kk]                   
Constraints_RHS_AT_PV_Max_Observed_Vars(kk)=Constraints_RHS_AT_PV_Max_Observed_Vars(kk)+1; %Now that we know the state, we increase the counter that we have seen it.

                        end %Ends loop over lines of dataset 1                            
                        Constraints_RHS_AT_PV_Max_Observed_Vars=Constraints_RHS_AT_PV_Max_Observed_Vars./nsamples;


                elseif all(ismember(PV_AND_Measured_Parents,vertices_Dataset2))==1
                    for qg=1:length(PV_AND_Measured_Parents)
                        Names_of_PV_AND_Measured_Parents(qg)=find(vertices_Dataset2==PV_AND_Measured_Parents(qg));
                    end
                    
%Right_Order_Names_of_Measured_Parents_and_PV=setdiff(Names_of_PV_AND_Measured_Parents,P_V); 
                        for nn=1:nsamples % we check every line of Dataset2                                                     
                                relevant_Part_Of_Line_nn=data_Two(nn,Names_of_PV_AND_Measured_Parents);
                            kk=1;
                                for nnn=1:length(PV_AND_Measured_Parents)%every partial line corresponds to some state. We next determine this state, called kk. 
kk=kk+2^(length(PV_AND_Measured_Parents)-nnn).*relevant_Part_Of_Line_nn(nnn);
                                end
                                %[PV_AND_Measured_Parents ,kk]                   
Constraints_RHS_AT_PV_Max_Observed_Vars(kk)=Constraints_RHS_AT_PV_Max_Observed_Vars(kk)+1; %Now that we know the state, we increase the counter that we have seen it.

                        end %Ends loop over lines of dataset 2                            
                        Constraints_RHS_AT_PV_Max_Observed_Vars=Constraints_RHS_AT_PV_Max_Observed_Vars./nsamples;

                    
                elseif all(ismember(PV_AND_Measured_Parents,vertices_Dataset3))==1
                    for qg=1:length(PV_AND_Measured_Parents)
                        Names_of_PV_AND_Measured_Parents(qg)=find(vertices_Dataset3==PV_AND_Measured_Parents(qg));
                    end
                    
%Right_Order_Names_of_Measured_Parents_and_PV=setdiff(Names_of_PV_AND_Measured_Parents,P_V); 
                        for nn=1:nsamples % we check every line of Dataset1                                                     
                                relevant_Part_Of_Line_nn=data_Three(nn,Names_of_PV_AND_Measured_Parents);
                            kk=1;
                                for nnn=1:length(PV_AND_Measured_Parents)%every partial line corresponds to some state. We next determine this state, called kk. 
kk=kk+2^(length(PV_AND_Measured_Parents)-nnn).*relevant_Part_Of_Line_nn(nnn);
                                end
                                %[PV_AND_Measured_Parents ,kk]                   
Constraints_RHS_AT_PV_Max_Observed_Vars(kk)=Constraints_RHS_AT_PV_Max_Observed_Vars(kk)+1; %Now that we know the state, we increase the counter that we have seen it.

                        end %Ends loop over lines of dataset 1                            
                        Constraints_RHS_AT_PV_Max_Observed_Vars=Constraints_RHS_AT_PV_Max_Observed_Vars./nsamples;

                end
              Constraints_RHS_at_PV=[Constraints_RHS_at_PV;Constraints_RHS_AT_PV_Max_Observed_Vars];          
        end% ends loop over set of jointly measured variables
        %P_V
        %length(LeftHandSide_of_Constraint_At_PV)
        %LeftHandSide_of_Constraint_At_PV
        %length(Constraints_RHS_at_PV)
        %Constraints_RHS_at_PV
        Constraints_At_PV=LeftHandSide_of_Constraint_At_PV-Constraints_RHS_at_PV;
        Constraints=[Constraints;Constraints_At_PV];%, or ;? I think the latter
    end% ends loop over all problematic vertices
end%ends loop over all vertices
%Constraints


%% Check whether brute force solution satisfies constraints....

% Test=[(Res(56)+Res(64))/(Res(56)+Res(55)+Res(64)+Res(63));...
% (Res(54)+Res(62))/(Res(54)+Res(53)+Res(62)+Res(61));...
% (Res(52)+Res(60))/(Res(52)+Res(51)+Res(60)+Res(59));...
% (Res(50)+Res(58))/(Res(50)+Res(49)+Res(58)+Res(57));...
% (Res(40)+Res(48))/(Res(40)+Res(39)+Res(48)+Res(47));...
% (Res(38)+Res(46))/(Res(38)+Res(37)+Res(46)+Res(45));...
% (Res(36)+Res(44))/(Res(36)+Res(35)+Res(44)+Res(43));...
% (Res(34)+Res(42))/(Res(34)+Res(33)+Res(42)+Res(41));...
% (Res(56-32)+Res(64-32))/(Res(56-32)+Res(55-32)+Res(64-32)+Res(63-32));...
% (Res(54-32)+Res(62-32))/(Res(54-32)+Res(53-32)+Res(62-32)+Res(61-32));...
% (Res(52-32)+Res(60-32))/(Res(52-32)+Res(51-32)+Res(60-32)+Res(59-32));...
% (Res(50-32)+Res(58-32))/(Res(50-32)+Res(49-32)+Res(58-32)+Res(57-32));...
% (Res(40-32)+Res(48-32))/(Res(40-32)+Res(39-32)+Res(48-32)+Res(47-32));...
% (Res(38-32)+Res(46-32))/(Res(38-32)+Res(37-32)+Res(46-32)+Res(45-32));...
% (Res(36-32)+Res(44-32))/(Res(36-32)+Res(35-32)+Res(44-32)+Res(43-32));...
% (Res(34-32)+Res(42-32))/(Res(34-32)+Res(33-32)+Res(42-32)+Res(41-32))]

% constraintsy(flipud(Test),data_One,data_Two,data_Three,nsamples,Y,vertices_Dataset1,vertices_Dataset2,vertices_Dataset3,N,Problematic_vertices,Final_Bayes_Net,Sum_Number_of_y_parameters_prior)

% Constraints =
%    1.0e-16 *
%     0.0087
%    -0.1388
%     0.1344
%          0
%    -0.0347
%    -0.0694
%    -0.0347
%    -0.2776
%    -0.0347
%    -0.1388
%     0.0347
%          0
%    -0.0347
%    -0.0694
%          0
%    -0.1388
%% I think this counts!